from river import datasets
from spotRiver.data.csvdataset import CSVDataset
from spotRiver.utils.data_conversion import convert_to_df


def data_selector(
    data_set,
) -> tuple:
    """
    Selects the river data set to be used.

    Args:
        data_set (str):
            Name of the data set to be used. Can be one of the following:
            - "Bananas"
            - "CreditCard"
            - "Elec2"
            - "Higgs"
            - "HTTP"
            - "Phishing"
            - "AirlinePassengers"
            - "Bikes"
            - "ChickWeights"
            - "Taxis"
            - "TrumpApproval"
            - "WaterFlow"
            - "WebTraffic"

    Returns:
        dataset (object):
            Data set to use. This is a dataset object from the river library.
        n_samples (int):
            Number of samples in the data set.

    Examples:
        >>> from spotPython.data.selector import data_selector
            dataset, n_samples = data_selector("Phishing")

    Notes:
        - The Higgs data set is very large and may take a long time to load.

    """
    # TODO: Check and update the number of samples for each data set.
    dataset = None
    if data_set == "Bananas":
        dataset = datasets.Bananas()
        n_samples = 5300
    elif data_set == "CreditCard":
        dataset = datasets.CreditCard()
        n_samples = 284_807
    elif data_set == "Elec2":
        dataset = datasets.Elec2()
        n_samples = 45_312
    elif data_set == "Higgs":
        dataset = datasets.Higgs()
        n_samples = 11_000_000
    elif data_set == "HTTP":
        dataset = datasets.HTTP()
        n_samples = 567_498
    elif data_set == "Phishing":
        dataset = datasets.Phishing()
        n_samples = 1250
    elif data_set == "AirlinePassengers":
        dataset = datasets.AirlinePassengers()
        n_samples = 144
    elif data_set == "Bikes":
        dataset = datasets.Bikes()
        n_samples = 182470
    elif data_set == "ChickWeights":
        dataset = datasets.ChickWeights()
        n_samples = 578
    elif data_set == "Taxis":
        dataset = datasets.Taxis()
        n_samples = 1458644
    elif data_set == "TrumpApproval":
        dataset = datasets.TrumpApproval()
        n_samples = 1_000
    elif data_set == "WaterFlow":
        dataset = datasets.WaterFlow()
        n_samples = 1_000
    else:
        raise ValueError(f"Data set '{data_set}' not found.")
    return dataset, n_samples


def get_train_test_from_data_set(df, n_total, test_size, target_column="y") -> tuple:
    """Converts a data set to a data frame with target column
        and splits it into training and test sets.

    Args:
        df (DataFrame):
            data set to be used.
        n_total (int):
            total number of samples to be used in the data set.
        test_size (float):
            percentage of the data set to be used as test set.
        target_column (str, optional):
            name of the target column. Defaults to "y".

    Returns:
        train:
            training data set.
        test:
            test data set.
        n_samples:
            total number of samples (train and test) in the data set.

    """
    n_features = len(df.columns) - 1
    df.columns = [f"x{i}" for i in range(1, n_features + 1)] + ["y"]
    df["y"] = df["y"].astype(int)
    # update n_samples to the actual number of samples in the data set,
    # because n_total might be smaller than n_samples which results in a smaller data set:
    test_size = float(test_size)
    n_samples = len(df)
    n_train = int((1.0 - test_size) * n_samples)
    train = df[:n_train]
    test = df[n_train:]
    return train, test, n_samples


def get_river_dataset_from_name(
    data_set_name,
    n_total=None,
    river_datasets=None,
) -> tuple:
    """Converts a data set name to a pandas DataFrame.

    Args:
        data_set_name (str):
            The name of the data set.
            If the data set name is not in river_datasets, the data set is assumed to be a CSV file.
        n_total (int):
            The number of samples to be used from the data set.
            If n_total is None or inf, the full data set is used.
            Defaults to None.
        river_datasets (list):
            A list of the available river data sets.
            If the data set name is not in river_datasets,
            the data set is assumed to be a CSV file.

    Returns:
        dataset (pd.DataFrame):
            The data set as a pandas DataFrame.
        n_samples (int):
            The number of samples in the data set.
    """
    if data_set_name.endswith(".csv"):
        dataset = CSVDataset(filename=data_set_name, directory="./userData/").data
        n_samples = dataset.shape[0]
    elif data_set_name in river_datasets:
        dataset, n_samples = data_selector(
            data_set=data_set_name,
        )
        dataset = convert_to_df(dataset, target_column="y", n_total=n_total)
    return dataset, n_samples



import pandas as pd
import numpy as np
from river import linear_model
from river import preprocessing
from sklearn.metrics import mean_absolute_error
from spotRiver.evaluation.eval_bml import eval_oml_horizon
from sklearn.metrics import accuracy_score
from river.forest import AMFClassifier
from river.datasets import Bananas
from spotRiver.hyperdict.river_hyper_dict import RiverHyperDict
from spotRiver.utils.data_conversion import convert_to_df
from spotPython.hyperparameters.values import add_core_model_to_fun_control
from spotPython.utils.init import fun_control_init
from spotPython.hyperparameters.values import modify_hyper_parameter_bounds
from spotPython.hyperparameters.values import get_one_core_model_from_X
from spotPython.hyperparameters.values import get_default_hyperparameters_as_array


def test_eval_oml_horizon():
    # create a sample model
    model = (
        preprocessing.StandardScaler() |
        linear_model.LinearRegression(intercept_lr=.5)
    )

    # create a sample train DataFrame
    train = pd.DataFrame({"x": np.arange(1, 11), "y": np.arange(2, 22, 2)})

    # create a sample test DataFrame
    test = pd.DataFrame({"x": np.arange(11, 111), "y": np.arange(22, 222, 2)})

    # set the target column
    target_column = "y"

    # set the horizon
    horizon = 5

    # set the metric
    metric = mean_absolute_error

    # evaluate the model
    res, preds = eval_oml_horizon(
        model=model,
        train=train,
        test=test,
        target_column=target_column,
        horizon=horizon,
        include_remainder=True,
        metric=metric,
        oml_grace_period=horizon,
    )

    # result should have one value for the initial model and one value for each horizon
    assert res.shape[0] == 1 + test.shape[0] // horizon
    # predictions  should be based on the test set only
    assert preds.shape == (test.shape[0], 3)


def test_eval_oml_horizon_with_default():
    PREFIX = "0000"
    fun_control = fun_control_init(
        PREFIX=PREFIX,
        TENSORBOARD_CLEAN=True)
    #fun_control
    horizon = 30
    n_samples = 5300
    n_train = 5000
    oml_grace_period = n_train
    dataset = Bananas(
    )
    target_column = "y"
    weights = np.array([- 1, 1/1000, 1/1000])*10_000.0
    df = convert_to_df(dataset, target_column=target_column, n_total=n_samples)
    df.columns = [f"x{i}" for i in range(1, dataset.n_features+1)] + ["y"]
    df["y"] = df["y"].astype(int)
    prep_model = preprocessing.StandardScaler()
    fun_control.update({"train":  df[:n_train],
                        "oml_grace_period": oml_grace_period,
                        "test":  df[n_train:],
                        "n_samples": n_samples,
                        "target_column": target_column,
                        "prep_model": prep_model,
                        "horizon": horizon,
                        "oml_grace_period": oml_grace_period,
                        "weights": weights,
                        "metric_sklearn": accuracy_score
                        })
    add_core_model_to_fun_control(core_model=AMFClassifier,
                                fun_control=fun_control,
                                hyper_dict=RiverHyperDict,
                                filename=None)
    modify_hyper_parameter_bounds(fun_control, "n_estimators", bounds=[2,20])
    modify_hyper_parameter_bounds(fun_control, "step", bounds=[0.5,2])

    X_start = get_default_hyperparameters_as_array(fun_control)
    model_default = get_one_core_model_from_X(X_start, fun_control, default=True)
    df_eval_default, df_true_default = eval_oml_horizon(
                        model=model_default,
                        train=fun_control["train"],
                        test=fun_control["test"],
                        target_column=fun_control["target_column"],
                        horizon=fun_control["horizon"],
                        oml_grace_period=fun_control["oml_grace_period"],
                        metric=fun_control["metric_sklearn"],
                    )
    assert df_eval_default.shape == ((n_samples - n_train) // horizon + 1,3)


import pandas as pd
from spotRiver.fun.hyperriver import HyperRiver

def test_compute_y():
    # create two Hyperriver objects
    hr = HyperRiver()
    hr_2 = HyperRiver(weights=[1.0, 2.0, 3.0])

    # create a sample evaluation DataFrame
    df_eval = pd.DataFrame({
        "Metric": [0.1, 0.2, 0.3],
        "CompTime (s)": [1.0, 2.0, 3.0],
        "Memory (MB)": [10.0, 20.0, 30.0]
    })

    # compute the objective function value
    y = hr.compute_y(df_eval)
    y_2 = hr_2.compute_y(df_eval)

    # check that the result is correct: calculate the means of the columns and multiply them with the weights
    assert y == 1.0 * df_eval["Metric"].mean() + 0.0 * df_eval["CompTime (s)"].mean() + 0.0 * df_eval["Memory (MB)"].mean()
    assert y_2 == 1.0 * df_eval["Metric"].mean() + 2.0 * df_eval["CompTime (s)"].mean() + 3.0 * df_eval["Memory (MB)"].mean()


import pandas as pd
import numpy as np


def corrplot(df: pd.DataFrame, numeric_only=True) -> None:
    """Function plots a graphical correlation matrix for each pair of columns in the dataframe.
        The function takes a DataFrame df as input and generates a graphical correlation matrix
        for each pair of columns in the dataframe.
        The upper triangle of the correlation matrix is masked out and set to NaN values.
        The resulting matrix is then styled and returned as a heatmap with colors
        ranging from blue (for negative correlations) to red (for positive correlations).

    Args:
        df (pd.DataFrame): Data to plot.
        numeric_only (bool, default True):
            Include only float, int or boolean data.
    Returns:
        (pd.DataFrame): A styled correlation matrix heatmap.

    Examples:
        >>> from spotRiver.plot.stats import corrplot
            import pandas as pd
            X = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6], 'C': [7, 8, 9]})
            corrplot(X)

    """

    # Compute the correlation matrix
    corr = df.corr(numeric_only=numeric_only)

    # Generate a mask for the upper triangle
    mask = np.zeros_like(corr, dtype=bool)
    mask[np.triu_indices_from(mask)] = True

    # Set values in the upper triangle to NaN
    corr[mask] = np.nan

    # Apply styling to the correlation matrix and return it as a heatmap
    return corr.style.background_gradient(cmap="coolwarm", axis=None, vmin=-1, vmax=1).highlight_null(color="#f1f1f1")



from river import datasets
import pandas as pd
from tabulate import tabulate
from sklearn.model_selection import train_test_split
from numpy.typing import ArrayLike
from math import inf


def convert_to_df(dataset: datasets.base.Dataset, target_column: str = "y", n_total: int = None) -> pd.DataFrame:
    """Converts a river dataset into a pandas DataFrame.

    Args:
        dataset (datasets.base.Dataset):
            The river dataset to be converted.
        target_column (str):
            The name of the target column in the resulting DataFrame.
            Defaults to "y".
        n_total (int, optional):
            The number of samples to be converted.
            If set to None or inf, the full dataset is converted.
            Defaults to None, i.e, the full dataset is converted.

    Returns:
        (pd.DataFrame): A pandas DataFrame representation of the given dataset.

    Examples:
        >>> dataset = datasets.TrumpApproval()
            target_column = "Approval"
            df = convert_to_df(dataset, target_column)
            df.rename(columns={
                'date': 'ordinal_date',
                'Gallup': 'gallup',
                'Ipsos': 'ipsos',
                'Morning Consult': 'morning_consult',
                'Rasmussen': 'rasmussen',
                'YouGov': 'you_gov'},
                inplace=True)
            # Split the data into train and test sets
            train = df[:500]
            test = df[500:]
    """
    data_dict = {key: [] for key in list(dataset.take(1))[0][0].keys()}
    data_dict[target_column] = []
    if n_total is None or n_total == inf:
        for x in dataset:
            for key, value in x[0].items():
                data_dict[key].append(value)
            data_dict[target_column].append(x[1])
    else:
        for x in dataset.take(n_total):
            for key, value in x[0].items():
                data_dict[key].append(value)
            data_dict[target_column].append(x[1])
    df = pd.DataFrame(data_dict)
    return df


def compare_two_tree_models(model1, model2, headers=["Parameter", "Default", "Spot"]):
    """Compares two tree models and returns a table of the differences.
    Args:
        model1 (Pipeline): A river model pipeline.
        model2 (Pipeline): A river model pipeline.
    Returns:
        (str): A table of the differences between the two models.
    """
    keys = model1[1].summary.keys()
    values1 = model1[1].summary.values()
    values2 = model2[1].summary.values()
    tbl = []
    for key, value1, value2 in zip(keys, values1, values2):
        tbl.append([key, value1, value2])
    return tabulate(tbl, headers=headers, numalign="right", tablefmt="github")


def rename_df_to_xy(df, target_column="y"):
    """Renames the columns of a DataFrame to x1, x2, ..., xn, y.

    Args:
        df (pd.DataFrame):
            The DataFrame to be renamed.
        target_column (str, optional):
            The name of the target column. Defaults to "y".

    Returns:
        (pd.DataFrame): The renamed DataFrame.

    Examples:
        >>> from spotRiver.utils.data_conversion import rename_df_to_xy
            df = pd.DataFrame({
            "feature1": [1, 2, 3],
            "feature2": [4, 5, 6],
            "target": [7, 8, 9]
        })
        >>> df = rename_df_to_xy(df, "target")
        >>> print(df)
           x1  x2  y
        0   1   4  7
        1   2   5  8
        2   3   6  9
    """
    n_features = len(df.columns) - 1
    df.columns = [f"x{i}" for i in range(1, n_features + 1)] + [target_column]
    return df


def split_df(
    dataset: pd.DataFrame, test_size: float, seed: int, stratify: ArrayLike, shuffle=True, target_type: str = None
) -> tuple:
    """
    Split a pandas DataFrame into a training and a test set.

    Args:
        dataset (pd.DataFrame):
            The input data set.
        test_size (float):
            The percentage of the data set to be used as test set.
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split.
            If int, represents the absolute number of test samples.
            If None, the value is set to the complement of the train size.
            If train_size is also None, it will be set to 0.25.
        target_type (str):
            The type of the target column. Can be "int", "float" or None.
            If None, the type of the target column is not changed.
            Otherwise, the target column is converted to the specified type.
        seed (int):
            The seed for the random number generator.
        stratify (ArrayLike):
            The array of target values.
        shuffle (bool, optional):
            Whether or not to shuffle the data before splitting. Defaults to True.

    Returns:
        tuple: The tuple (train, test, n_samples).

    Examples:
        >>> from spotRiver.utils.data_conversion import split_df
            df = pd.DataFrame({
            "feature1": [1, 2, 3],
            "feature2": [4, 5, 6],
            "target": [7, 8, 9]})
            train, test, n_samples = split_df(df, 0.2, "int", 42)

    """
    # Rename the columns of a DataFrame to x1, x2, ..., xn, y.
    # From now on we assume that the target column is called "y":
    df = rename_df_to_xy(df=dataset, target_column="y")
    if target_type == "float":
        df["y"] = df["y"].astype(float)
    elif target_type == "int":
        df["y"] = df["y"].astype(int)
    else:
        pass
    target_column = "y"
    # split the data set into a training and a test set,
    # where the test set is a percentage of the data set given as test_size:
    X = df.drop(columns=[target_column])
    Y = df[target_column]
    # Split the data into training and test sets
    # test_size is the percentage of the data that should be held over for testing
    # random_state is a seed for the random number generator to make your train and test splits reproducible
    train_features, test_features, train_target, test_target = train_test_split(
        X, Y, test_size=test_size, random_state=seed, shuffle=shuffle, stratify=stratify
    )
    # combine the training features and the training target into a training DataFrame
    train = pd.concat([train_features, train_target], axis=1)
    test = pd.concat([test_features, test_target], axis=1)
    n_samples = train.shape[0] + test.shape[0]
    return train, test, n_samples



from spotRiver.data.selector import data_selector
from spotRiver.utils.data_conversion import convert_to_df

def test_data_selector():
    data_set_values = [
        "Bananas",
        "Elec2",
        "HTTP",
        "Phishing",
    ]
    for i in range(len(data_set_values)):
        data_set = data_set_values[i]
        dataset, n_samples = data_selector(
            data_set=data_set
        )
        df = convert_to_df(dataset, target_column="y", n_total=None)
        assert df.shape[0] == n_samples


from river import time_series
from river import compose
from river import linear_model
from river import optim
from river import preprocessing
from river import metrics
from numpy.random import default_rng
import numpy as np
from numpy import array

from spotRiver.utils.features import get_weekday_distances
from spotRiver.utils.features import get_ordinal_date
from spotRiver.utils.features import get_month_distances
from spotRiver.utils.features import get_hour_distances
from spotRiver.evaluation.eval_oml import fun_eval_oml_iter_progressive
from spotRiver.evaluation.eval_oml import eval_oml_iter_progressive
from spotRiver.evaluation.eval_bml import eval_oml_horizon
from spotRiver.evaluation.eval_nowcast import eval_nowcast_model

from spotPython.hyperparameters.values import (
    assign_values,
)
from spotPython.hyperparameters.prepare import (
    get_dict_with_levels_and_types,
    get_one_config_from_var_dict,
    iterate_dict_values,
    convert_keys,
)

from spotPython.utils.transform import transform_hyper_parameter_values

import logging
import statistics
from sklearn.metrics import mean_absolute_error

logger = logging.getLogger(__name__)
# configure the handler and formatter as needed
py_handler = logging.FileHandler(f"{__name__}.log", mode="w")
py_formatter = logging.Formatter("%(name)s %(asctime)s %(levelname)s %(message)s")
# add formatter to the handler
py_handler.setFormatter(py_formatter)
# add handler to the logger
logger.addHandler(py_handler)


class HyperRiver:
    """
    Hyperparameter Tuning for River.

    Args:
        seed (int): seed.
            See [Numpy Random Sampling](https://numpy.org/doc/stable/reference/random/index.html#random-quick-start)

    """

    def __init__(self, seed=126, log_level=50):
        self.seed = seed
        self.rng = default_rng(seed=self.seed)
        self.fun_control = {
            "seed": None,
            "data": None,
            "step": 10_000,
            "horizon": None,
            "grace_period": None,
            "metric": metrics.MAE(),
            "metric_sklearn": mean_absolute_error,
            "weights": array([1, 0, 0]),
            "weight_coeff": 0.0,
            "log_level": log_level,
            "var_name": [],
            "var_type": [],
            "prep_model": None,
        }
        self.log_level = self.fun_control["log_level"]
        logger.setLevel(self.log_level)
        logger.info(f"Starting the logger at level {self.log_level} for module {__name__}:")

    # def get_month_distances(x):
    #     return {
    #         calendar.month_name[month]: math.exp(-(x['month'].month - month) ** 2)
    #         for month in range(1, 13)
    #     }

    # def get_ordinal_date(x):
    #     return {'ordinal_date': x['month'].toordinal()}
    def fun_nowcasting(self, X, fun_control=None):
        """Hyperparameter Tuning of the nowcasting model.

        Returns:
            (float): objective function value. Mean of the MAEs of the predicted values.
        """
        self.fun_control.update(fun_control)

        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 5:
            raise Exception
        lr = X[:, 0]
        intercept_lr = X[:, 1]
        hour = X[:, 2]
        weekday = X[:, 3]
        month = X[:, 4]

        z_res = np.array([], dtype=float)
        for i in range(X.shape[0]):
            h_i = int(hour[i])
            w_i = int(weekday[i])
            m_i = int(month[i])
            # baseline:
            extract_features = compose.TransformerUnion(get_ordinal_date)
            if h_i:
                extract_features = compose.TransformerUnion(get_ordinal_date, get_hour_distances)
            if w_i:
                extract_features = compose.TransformerUnion(extract_features, get_weekday_distances)
            if m_i:
                extract_features = compose.TransformerUnion(extract_features, get_month_distances)
            model = compose.Pipeline(
                ("features", extract_features),
                ("scale", preprocessing.StandardScaler()),
                (
                    "lin_reg",
                    linear_model.LinearRegression(
                        intercept_init=0, optimizer=optim.SGD(float(lr[i])), intercept_lr=float(intercept_lr[i])
                    ),
                ),
            )
            # eval:
            dates, metric, y_trues, y_preds = eval_nowcast_model(
                model, dataset=self.fun_control["data"], time_interval="Time"
            )
            z = metric.get()
            z_res = np.append(z_res, z)
        return z_res

    def fun_snarimax(self, X, fun_control=None):
        """Hyperparameter Tuning of the SNARIMAX model.
            SNARIMAX stands for (S)easonal (N)on-linear (A)uto(R)egressive (I)ntegrated (M)oving-(A)verage with
            e(X)ogenous inputs model.
        This model generalizes many established time series models in a single interface that can be
        trained online. It assumes that the provided training data is ordered in time and is uniformly spaced.
        It is made up of the following components:
        - S (Seasonal)
        - N (Non-linear): Any online regression model can be used, not necessarily a linear regression
            as is done in textbooks.
        - AR (Autoregressive): Lags of the target variable are used as features.
        - I (Integrated): The model can be fitted on a differenced version of a time series. In this
            context, integration is the reverse of differencing.
        - MA (Moving average): Lags of the errors are used as features.
        - X (Exogenous): Users can provide additional features. Care has to be taken to include
            features that will be available both at training and prediction time.

        Each of these components can be switched on and off by specifying the appropriate parameters.
        Classical time series models such as AR, MA, ARMA, and ARIMA can thus be seen as special
        parametrizations of the SNARIMAX model.

        This model is tailored for time series that are homoskedastic. In other words, it might not
        work well if the variance of the time series varies widely along time.

        Parameters of the hyperparameter vector:

            `p`: Order of the autoregressive part. This is the number of past target values that will be
                included as features.
            `d`: Differencing order.
            `q`: Order of the moving average part. This is the number of past error terms that will be included
                as features.
            `m`: Season length used for extracting seasonal features. If you believe your data has a seasonal
                pattern, then set this accordingly. For instance, if the data seems to exhibit a yearly seasonality,
                and that your data is spaced by month, then you should set this to `12`.
                Note that for this parameter to have any impact you should also set at least
                one of the `p`, `d`, and `q` parameters.
            `sp`:  Seasonal order of the autoregressive part. This is the number of past target values that will
                be included as features.
            `sd`: Seasonal differencing order.
            `sq`: Seasonal order of the moving average part. This is the number of past error terms that will be
                included as features.
            `lr` (float):
                learn rate of the linear regression model. A river `preprocessing.StandardScaler`
                piped with a river `linear_model.LinearRegression` will be used.
            `intercept_lr` (float): intercept of the the linear regression model. A river `preprocessing.StandardScaler`
                piped with a river `linear_model.LinearRegression` will be used.
            `hour` (bool): If `True`, an hourly component is added.
            `weekdy` (bool): If `True`, an weekday component is added.
            `month` (bool): If `True`, an monthly component is added.

        Args:
            X (array):
                Seven hyperparameters to be optimized. Here:

                `p` (int):
                    Order of the autoregressive part.
                    This is the number of past target values that will be included as features.

                `d` (int):
                    Differencing order.

                `q` (int):
                    Order of the moving average part.
                    This is the number of past error terms that will be included as features.

                `m` (int):
                    Season length used for extracting seasonal features.
                    If you believe your data has a seasonal pattern, then set this accordingly.
                    For instance, if the data seems to exhibit a yearly seasonality,
                    and that your data is spaced by month, then you should set this to `12`.
                    Note that for this parameter to have any impact you should also set
                    at least one of the `p`, `d`, and `q` parameters.

                `sp` (int):
                    Seasonal order of the autoregressive part.
                    This is the number of past target values that will be included as features.

                `sd` (int):
                    Seasonal differencing order.

                `sq`(int):
                    Seasonal order of the moving average part.
                    This is the number of past error terms that will be included as features.

                `lr` (float):
                    learn rate of the linear regression model. A river `preprocessing.StandardScaler`
                    piped with a river `linear_model.LinearRegression` will be used.

                `intercept_lr` (float): intercept of the the linear regression model.
                    A river `preprocessing.StandardScaler` piped with a river `linear_model.LinearRegression`
                    will be used.

                `hour` (bool): If `True`, an hourly component is added.

                `weekday` (bool): If `True`, an weekday component is added.

                `month` (bool): If `True`, an monthly component is added.

            fun_control (dict):
                parameter that are not optimized, e.g., `horizon`. Commonly
                referred to as "design of experiments" parameters:

                1. `horizon`: (int)

                2. `grace_period`: (int) Initial period during which the metric is not updated.
                    This is to fairly evaluate models which need a warming up period to start
                    producing meaningful forecasts.
                    The value of this parameter is equal to the `horizon` by default.

                3. `data`: dataset. Default `AirlinePassengers`.

        Returns:
            (float): objective function value. Mean of the MAEs of the predicted values.
        """
        self.fun_control.update(fun_control)

        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 12:
            raise Exception
        p = X[:, 0]
        d = X[:, 1]
        q = X[:, 2]
        m = X[:, 3]
        sp = X[:, 4]
        sd = X[:, 5]
        sq = X[:, 6]
        lr = X[:, 7]
        intercept_lr = X[:, 8]
        hour = X[:, 9]
        weekday = X[:, 10]
        month = X[:, 11]

        # TODO:
        # horizon = fun_control["horizon"]
        # future = [
        #   {"month": dt.date(year=1961, month=m, day=1)} for m in range(1, horizon + 1)
        # ]
        z_res = np.array([], dtype=float)
        for i in range(X.shape[0]):
            h_i = int(hour[i])
            w_i = int(weekday[i])
            m_i = int(month[i])
            # baseline:
            extract_features = compose.TransformerUnion(get_ordinal_date)
            if h_i:
                extract_features = compose.TransformerUnion(get_ordinal_date, get_hour_distances)
            if w_i:
                extract_features = compose.TransformerUnion(extract_features, get_weekday_distances)
            if m_i:
                extract_features = compose.TransformerUnion(extract_features, get_month_distances)
            model = compose.Pipeline(
                extract_features,
                time_series.SNARIMAX(
                    p=int(p[i]),
                    d=int(d[i]),
                    q=int(q[i]),
                    m=int(m[i]),
                    sp=int(sp[i]),
                    sd=int(sd[i]),
                    sq=int(sq[i]),
                    regressor=compose.Pipeline(
                        preprocessing.StandardScaler(),
                        linear_model.LinearRegression(
                            intercept_init=0,
                            optimizer=optim.SGD(float(lr[i])),
                            intercept_lr=float(intercept_lr[i]),
                        ),
                    ),
                ),
            )
            # eval:
            res = time_series.evaluate(
                self.fun_control["data"],
                model,
                metric=self.fun_control["metric"],
                horizon=self.fun_control["horizon"],
                agg_func=statistics.mean,
            )
            z = res.get()
            z_res = np.append(z_res, z)
        return z_res

    def fun_hw(self, X, fun_control=None):
        """Hyperparameter Tuning of the HoltWinters model.

            Holt-Winters forecaster. This is a standard implementation of the Holt-Winters forecasting method.
            Certain parameterizations result in special cases, such as simple exponential smoothing.

        Args:
            X (array): five hyperparameters. The parameters of the hyperparameter vector are:

                1. `alpha`: Smoothing parameter for the level.
                2. `beta`: Smoothing parameter for the trend.
                3. `gamma`: Smoothing parameter for the seasonality.
                4. `seasonality`: The number of periods in a season.
                    For instance, this should be 4 for quarterly data, and 12 for yearly data.
                5. `multiplicative`: Whether or not to use a multiplicative formulation.

            fun_control (dict): Parameters that are are not tuned:

                1. `horizon`: (int)
                2. `grace_period`: (int) Initial period during which the metric is not updated.
                    This is to fairly evaluate models which need a warming up period to start
                    producing meaningful forecasts.
                    The value of this parameter is equal to the `horizon` by default.
                2. `data`: dataset. Default `AirlinePassengers`.

        Returns:
            (float): objective function value. Mean of the MAEs of the predicted values.
        """
        self.fun_control.update(fun_control)
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 5:
            raise Exception

        alpha = X[:, 0]
        beta = X[:, 1]
        gamma = X[:, 2]
        seasonality = X[:, 3]
        multiplicative = X[:, 4]
        z_res = np.array([], dtype=float)
        for i in range(X.shape[0]):
            model = time_series.HoltWinters(
                alpha=alpha[i],
                beta=beta[i],
                gamma=gamma[i],
                seasonality=int(seasonality[i]),
                multiplicative=int(multiplicative[i]),
            )
            res = time_series.evaluate(
                self.fun_control["data"],
                model,
                metric=self.fun_control["metric"],
                horizon=self.fun_control["horizon"],
                grace_period=self.fun_control["grace_period"],
                agg_func=statistics.mean,
            )
            z = res.get()
            z_res = np.append(z_res, z)
        return z_res

    # def fun_HTR_iter_progressive(self, X, fun_control=None):
    #     """Hyperparameter Tuning of HTR model.
    #     See: https://riverml.xyz/0.15.0/api/tree/HoeffdingTreeRegressor/
    #     Parameters
    #     ----------
    #     grace_period
    #         Number of instances a leaf should observe between split attempts.
    #     max_depth
    #         The maximum depth a tree can reach. If `None`, the tree will grow indefinitely.
    #     delta
    #         Significance level to calculate the Hoeffding bound. The significance level is given by
    #         `1 - delta`. Values closer to zero imply longer split decision delays.
    #     tau
    #         Threshold below which a split will be forced to break ties.
    #     leaf_prediction
    #         Prediction mechanism used at leafs. NOTE: order differs from the order in river!</br>
    #         - 'mean' - Target mean</br>
    #         - 'adaptive' - Chooses between 'mean' and 'model' dynamically</br>
    #         - 'model' - Uses the model defined in `leaf_model`</br>
    #     NOT IMPLEMENTED: leaf_model
    #         The regression model used to provide responses if `leaf_prediction='model'`. If not
    #         provided an instance of `river.linear_model.LinearRegression` with the default
    #         hyperparameters is used.
    #     model_selector_decay
    #         The exponential decaying factor applied to the learning models' squared errors, that
    #         are monitored if `leaf_prediction='adaptive'`. Must be between `0` and `1`. The closer
    #         to `1`, the more importance is going to be given to past observations. On the other hand,
    #         if its value approaches `0`, the recent observed errors are going to have more influence
    #         on the final decision.
    #     nominal_attributes
    #         List of Nominal attributes identifiers. If empty, then assume that all numeric attributes
    #         should be treated as continuous.
    #     splitter
    #         The Splitter or Attribute Observer (AO) used to monitor the class statistics of numeric
    #         features and perform splits. Splitters are available in the `tree.splitter` module.
    #         Different splitters are available for classification and regression tasks. Classification
    #         and regression splitters can be distinguished by their property `is_target_class`.
    #         This is an advanced option. Special care must be taken when choosing different splitters.
    #         By default, `tree.splitter.TEBSTSplitter` is used if `splitter` is `None`.
    #     min_samples_split
    #         The minimum number of samples every branch resulting from a split candidate must have
    #         to be considered valid.
    #     binary_split
    #         If True, only allow binary splits.
    #     max_size
    #         The max size of the tree, in Megabytes (MB).
    #     memory_estimate_period
    #         Interval (number of processed instances) between memory consumption checks.
    #     stop_mem_management
    #         If True, stop growing as soon as memory limit is hit.
    #     remove_poor_attrs
    #         If True, disable poor attributes to reduce memory usage.
    #     merit_preprune
    #         If True, enable merit-based tree pre-pruning.

    #     fun_control
    #         Parameters that are are not tuned:
    #             1. `horizon`: (int)
    #             2. `grace_period`: (int) Initial period during which the metric is not updated.
    #                     This is to fairly evaluate models which need a warming up period to start
    #                     producing meaningful forecasts.
    #                     The value of this parameter is equal to the `horizon` by default.
    #             3. `data`: dataset. Default `AirlinePassengers`.

    #     Returns
    #     -------
    #     (float): objective function value. Mean of the MAEs of the predicted values.
    #     """
    #     self.fun_control.update(fun_control)
    #     try:
    #         X.shape[1]
    #     except ValueError:
    #         X = np.array([X])
    #     if X.shape[1] != 11:
    #         raise Exception
    #     grace_period = X[:, 0]
    #     max_depth = X[:, 1]
    #     delta = X[:, 2]
    #     tau = X[:, 3]
    #     leaf_prediction = X[:, 4]
    #     leaf_model = X[:, 5]
    #     model_selector_decay = X[:, 6]
    #     splitter = X[:, 7]
    #     min_samples_split = X[:, 8]
    #     binary_split = X[:, 9]
    #     max_size = X[:, 10]
    #     z_res = np.array([], dtype=float)
    #     dataset_list = self.fun_control["data"]
    #     for i in range(X.shape[0]):
    #         num = compose.SelectType(numbers.Number) | preprocessing.StandardScaler()
    #         cat = compose.SelectType(str) | preprocessing.FeatureHasher(n_features=1000, seed=1)
    #         try:
    #             res = eval_oml_iter_progressive(
    #                 dataset=dataset_list,
    #                 step=self.fun_control["step"],
    #                 log_level=self.fun_control["log_level"],
    #                 metric=fun_control["metric"],
    #                 weight_coeff=fun_control["weight_coeff"],
    #                 models={
    #                     "HTR": (
    #                         (num + cat)
    #                         | tree.HoeffdingTreeRegressor(
    #                             grace_period=int(grace_period[i]),
    #                             max_depth=transform_power_10(int(max_depth[i])),
    #                             delta=float(delta[i]),
    #                             tau=float(tau[i]),
    #                             leaf_prediction=select_leaf_prediction(int(leaf_prediction[i])),
    #                             leaf_model=select_leaf_model(int(leaf_model[i])),
    #                             model_selector_decay=float(model_selector_decay[i]),
    #                             splitter=select_splitter(int(splitter[i])),
    #                             min_samples_split=int(min_samples_split[i]),
    #                             binary_split=int(binary_split[i]),
    #                             max_size=float(max_size[i]),
    #                         )
    #                     ),
    #                 },
    #             )
    #             logger.debug("res from eval_oml_iter_progressive: %s", res)
    #             y = fun_eval_oml_iter_progressive(res, metric=None, weights=self.fun_control["weights"])
    #         except Exception as err:
    #             y = np.nan
    #             print(f"Error in fun(). Call to evaluate failed. {err=}, {type(err)=}")
    #             print(f"Setting y to {y:.2f}.")
    #         z_res = np.append(z_res, y / self.fun_control["n_samples"])
    #     return z_res

    def fun_oml_iter_progressive(self, X, fun_control=None):
        """Hyperparameter Tuning of an arbitrary model.
        Returns
        -------
        (float): objective function value. Mean of the MAEs of the predicted values.
        """
        self.fun_control.update(fun_control)
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != len(self.fun_control["var_name"]):
            raise Exception
        var_dict = assign_values(X, self.fun_control["var_name"])
        z_res = np.array([], dtype=float)
        dataset_list = self.fun_control["data"]
        for values in iterate_dict_values(var_dict):
            values = convert_keys(values, self.fun_control["var_type"])
            print(values)
            values = get_dict_with_levels_and_types(fun_control=self.fun_control, v=values)
            values = transform_hyper_parameter_values(fun_control=self.fun_control, hyper_parameter_values=values)
            print(values)
            model = compose.Pipeline(self.fun_control["prep_model"], self.fun_control["core_model"](**values))
            try:
                res = eval_oml_iter_progressive(
                    dataset=dataset_list,
                    step=self.fun_control["step"],
                    log_level=self.fun_control["log_level"],
                    metric=fun_control["metric"],
                    weight_coeff=fun_control["weight_coeff"],
                    models={
                        self.fun_control["model_name"]: (model),
                    },
                )
                logger.debug("res from eval_oml_iter_progressive: %s", res)
                y = fun_eval_oml_iter_progressive(res, metric=None, weights=self.fun_control["weights"])
            except Exception as err:
                y = np.nan
                print(f"Error in fun(). Call to evaluate failed. {err=}, {type(err)=}")
                print(f"Setting y to {y:.2f}.")
            z_res = np.append(z_res, y / self.fun_control["n_samples"])
        return z_res

    def compute_y(self, df_eval):
        """Compute the objective function value.
        Args:
            df_eval (pd.DataFrame): DataFrame with the evaluation results.
        Returns:
            (float): objective function value. Mean of the MAEs of the predicted values.
        Examples:
            >>> df_eval = pd.DataFrame( [[1, 2, 3], [4, 5, 6]], columns=['Metric', 'CompTime (s)', 'Memory (MB)'])
            >>> weights = [1, 1, 1]
            >>> compute_y(df_eval, weights)
            4.0
        """
        # take the mean of the MAEs/ACCs of the predicted values and ignore the NaN values
        df_eval = df_eval.dropna()
        y_error = df_eval["Metric"].mean()
        logger.debug("y_error from eval_oml_horizon: %s", y_error)
        y_r_time = df_eval["CompTime (s)"].mean()
        logger.debug("y_r_time from eval_oml_horizon: %s", y_r_time)
        y_memory = df_eval["Memory (MB)"].mean()
        logger.debug("y_memory from eval_oml_horizon: %s", y_memory)
        weights = self.fun_control["weights"]
        logger.debug("weights from eval_oml_horizon: %s", weights)
        y = weights[0] * y_error + weights[1] * y_r_time + weights[2] * y_memory
        logger.debug("weighted res from eval_oml_horizon: %s", y)
        return y

    # def fun_oml_horizon_old(self, X, fun_control=None, return_model=False, return_df=False):
    #     """Hyperparameter Tuning of an arbitrary model.
    #     Returns
    #     -------
    #     (float): objective function value. Mean of the MAEs of the predicted values.
    #     """
    #     self.fun_control.update(fun_control)
    #     weights = self.fun_control["weights"]
    #     if len(weights) != 3:
    #         raise ValueError("The weights array must be of length 3.")
    #     try:
    #         X.shape[1]
    #     except ValueError:
    #         X = np.array([X])
    #     if X.shape[1] != len(self.fun_control["var_name"]):
    #         raise Exception
    #     var_dict = assign_values(X, self.fun_control["var_name"])
    #     z_res = np.array([], dtype=float)
    #     for values in iterate_dict_values(var_dict):
    #         values = convert_keys(values, self.fun_control["var_type"])
    #         values = get_dict_with_levels_and_types(fun_control=self.fun_control, v=values)
    #         values = transform_hyper_parameter_values(fun_control=self.fun_control, hyper_parameter_values=values)
    #         model = compose.Pipeline(self.fun_control["prep_model"], self.fun_control["core_model"](**values))
    #         if return_model:
    #             return model
    #         try:
    #             df_eval, df_preds = eval_oml_horizon(
    #                 model=model,
    #                 train=self.fun_control["train"],
    #                 test=self.fun_control["test"],
    #                 target_column=self.fun_control["target_column"],
    #                 horizon=self.fun_control["horizon"],
    #                 oml_grace_period=self.fun_control["oml_grace_period"],
    #                 metric=self.fun_control["metric_sklearn"],
    #             )
    #         except Exception as err:
    #             print(f"Error in fun_oml_horizon(). Call to eval_oml_horizon failed. {err=}, {type(err)=}")
    #         if return_df:
    #             return df_eval, df_preds
    #         try:
    #             y = self.compute_y(df_eval, weights)
    #         except Exception as err:
    #             y = np.nan
    #             print(f"Error in fun(). Call to evaluate failed. {err=}, {type(err)=}")
    #             print(f"Setting y to {y:.2f}.")
    #         z_res = np.append(z_res, y / self.fun_control["n_samples"])
    #     return z_res

    def check_weights(self):
        if len(self.fun_control["weights"]) != 3:
            raise ValueError("The weights array must be of length 3.")

    def check_X_shape(self, X):
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != len(self.fun_control["var_name"]):
            raise Exception

    def evaluate_model(self, model, fun_control):
        try:
            df_eval, df_preds = eval_oml_horizon(
                model=model,
                train=fun_control["train"],
                test=fun_control["test"],
                target_column=fun_control["target_column"],
                horizon=fun_control["horizon"],
                oml_grace_period=fun_control["oml_grace_period"],
                metric=fun_control["metric_sklearn"],
            )
        except Exception as err:
            print(f"Error in fun_oml_horizon(). Call to eval_oml_horizon failed. {err=}, {type(err)=}")
        return df_eval, df_preds

    def get_river_df_eval_preds(self, model):
        try:
            df_eval, df_preds = self.evaluate_model(model, self.fun_control)
        except Exception as err:
            print(f"Error in get_river_df_eval_preds(). Call to evaluate_model failed. {err=}, {type(err)=}")
            print("Setting df_eval and df.preds to np.nan")
            df_eval = np.nan
            df_preds = np.nan
        return df_eval, df_preds

    def fun_oml_horizon_old(self, X, fun_control=None):
        z_res = np.array([], dtype=float)
        self.fun_control.update(fun_control)
        self.check_weights()
        self.check_X_shape(X)
        var_dict = assign_values(X, self.fun_control["var_name"])
        for config in get_one_config_from_var_dict(var_dict, self.fun_control):
            if self.fun_control["prep_model"] is not None:
                model = compose.Pipeline(self.fun_control["prep_model"], self.fun_control["core_model"](**config))
            else:
                model = self.fun_control["core_model"](**config)
            try:
                df_eval, _ = self.evaluate_model(model, self.fun_control)
            except Exception as err:
                df_eval = np.nan
                print(f"Error in fun(). Call to evaluate failed. {err=}, {type(err)=}")
                print("Setting df_eval to np.nan.")
            try:
                y = self.compute_y(df_eval)
            except Exception as err:
                y = np.nan
                print(f"Error in fun(). Call to compute_y failed. {err=}, {type(err)=}")
                print("Setting y to np.nan.")
            z_res = np.append(z_res, y / self.fun_control["n_samples"])
        return z_res

    def fun_oml_horizon(self, X, fun_control=None):
        """
        This function calculates the horizon for a given set of data X and control parameters.

        :param X: numpy array of data
        :param fun_control: dictionary of control parameters
        :return: numpy array of horizon values
        """
        self.fun_control.update(fun_control)
        self.check_weights()
        self.check_X_shape(X)
        var_dict = assign_values(X, self.fun_control["var_name"])
        z_res = []
        for config in get_one_config_from_var_dict(var_dict, self.fun_control):
            if self.fun_control["prep_model"] is not None:
                model = compose.Pipeline(self.fun_control["prep_model"], self.fun_control["core_model"](**config))
            else:
                model = self.fun_control["core_model"](**config)
            try:
                df_eval, _ = self.evaluate_model(model, self.fun_control)
                y = self.compute_y(df_eval)
            except Exception as err:
                y = np.nan
                print(f"Error in fun(). Call to evaluate or compute_y failed. {err=}, {type(err)=}")
                print("Setting y to np.nan.")
            z_res.append(y / self.fun_control["n_samples"])
        return np.array(z_res)



from sklearn.datasets import fetch_openml


def get_bike_sharing_data(train_size=0.6):
    """
    Fetches the Bike Sharing Demand dataset from OpenML and splits it into training and test sets.

    Args:
        train_size (float):
            The proportion of the dataset to include in the training set. Default value: 0.6

    Returns:
        (tuple): tuple containing:
            df (pd.DataFrame): The full dataset.
            train (pd.DataFrame): The training set.
            test (pd.DataFrame): The test set.

    Examples:
        >>> from spotRiver.data.bike_sharing import get_bike_sharing_data
        >>> df, train, test = get_bike_sharing_data(train_size=0.6)
    """

    bike_sharing = fetch_openml("Bike_Sharing_Demand", version=2, as_frame=True, parser="pandas")
    df = bike_sharing.frame
    # Normalize the count column
    df["count"] = df["count"] / df["count"].max()
    # Replace heavy_rain with rain in the weather column
    df["weather"].replace(to_replace="heavy_rain", value="rain", inplace=True)
    n = df.shape[0]
    # Calculate the number of rows in the training set
    k = int(n * train_size)
    # Split the data into training and test sets
    train = df[0:k]
    test = df[k:n]
    return df, train, test



import pandas as pd
import numpy as np
from spotRiver.data.opm import fetch_opm
from sklearn.impute import SimpleImputer


def impute_opm(
    include_categorical: bool = False,
    data_home: str = "data",
    strategy: str = "most_frequent",
    columns: list[str] = ["lat", "lon"],
    archive_name: str = "opm_cat.csv",
    path_or_buf: str = "opm_cat.zip",
    write_csv: bool = True,
    return_df: bool = False,
) -> pd.DataFrame:
    """Impute missing values in OPM dataset.

    Args:
        include_categorical: Whether to include categorical features. Default is False.
        data_home: The directory to use as a data store. Default is "data".
        strategy: The imputation strategy to use. Can be one of "mean", "median", "most_frequent", or "constant". Default is "most_frequent".
        columns: A list of column names to impute. If None, impute all columns. Default is ["lat", "lon"].
        archive_name: The name of the archive file to write. Default is "opm_cat.csv".
        path_or_buf: The file path or buffer to write. Default is "opm_cat.zip".
        write_csv: Whether to write the imputed data to a CSV file. Default is True.
        return_df: Whether to return the imputed data as a DataFrame. Default is False.

    Returns:
        If `return_df` is True, returns a pandas DataFrame containing the imputed data.
    """
    # Validate input parameters
    valid_strategies = ["mean", "median", "most_frequent", "constant"]
    if strategy not in valid_strategies:
        raise ValueError(f"Invalid strategy: {strategy}. Must be one of {valid_strategies}.")
    # Fetch and concatenate data
    X, y = fetch_opm(include_categorical=include_categorical, data_home=data_home, return_X_y=True)
    df = pd.concat([X, y], axis=1)
    # Impute missing values
    imp = SimpleImputer(missing_values=np.nan, strategy=strategy)
    if columns is None:
        # Impute all columns
        df[:] = imp.fit_transform(df)
    else:
        # Impute only specified columns
        for col in columns:
            if col not in df.columns:
                raise ValueError(f"Invalid column: {col}. Not in dataframe.")
            df[col] = imp.fit_transform(np.array(df[col]).reshape(-1, 1))
    # Write csv file if requested
    if write_csv:
        compression_opts = dict(method="zip", archive_name=archive_name)
        df.to_csv(path_or_buf, index=False, compression=compression_opts)
    # Return dataframe if requested
    if return_df:
        return df



import pandas as pd
import pathlib


class CSVDataset:
    """
    A Dataset for handling CSV data.

    Args:
        filename (str): The path to the CSV file. Defaults to "data.csv".
        directory (str): The path to the directory where the CSV file is stored. Defaults to None.
        feature_type (dtype): The data type of the features. Defaults to float.
        target_column (str): The name of the target column. Defaults to "y".
        target_type (dtype): The data type of the targets. Defaults to float.
        train (bool): Whether the dataset is for training or not. Defaults to True.
        rmNA (bool): Whether to remove rows with NA values or not. Defaults to True.
        dropId (bool): Whether to drop the "id" column or not. Defaults to False.
        **desc (Any): Additional keyword arguments.

    Examples:
        >>> from spotRiver.data.csvdataset import CSVDataset
            dataset = CSVDataset(csv_file='data.csv', target_column='prognosis', feature_type=int)
            print(dataset.data.shape)
    """

    def __init__(
        self,
        filename: str = None,
        directory: None = None,
        feature_type: type = float,
        target_column: str = "y",
        target_type: type = float,
        train: bool = True,
        rmNA=True,
        dropId=False,
        **desc,
    ) -> None:
        # super().__init__()
        self.filename = filename
        self.directory = directory
        self.feature_type = feature_type
        self.target_type = target_type
        self.target_column = target_column
        self.train = train
        self.rmNA = rmNA
        self.dropId = dropId
        self.data = self._load_data()

    @property
    def path(self):
        if self.directory:
            return pathlib.Path(self.directory).joinpath(self.filename)
        return pathlib.Path(__file__).parent.joinpath(self.filename)

    @property
    def _repr_content(self):
        content = super()._repr_content
        content["Path"] = str(self.path)
        return content

    def _load_data(self) -> tuple:
        # print(f"Loading data from {self.path}")
        df = pd.read_csv(self.path, index_col=False)
        # rm rows with NA
        if self.rmNA:
            df = df.dropna()
        if self.dropId:
            df = df.drop(columns=["id"])
        return df



import matplotlib.pyplot as plt
from river.evaluate import iter_progressive_val_score
from spotPython.utils.progress import progress_bar
from numpy import mean
from numpy import zeros
from numpy import array


def eval_oml_iter_progressive(dataset, metric, models, step=100, weight_coeff=0.0, log_level=50):
    """Evaluate OML Models on Streaming Data

    This function evaluates one or more OML models on a streaming dataset. The evaluation
    is done iteratively, and the models are tested every `step` iterations. The results
    are returned as a dictionary of metrics and their values.

    Args:
        dataset (list or river.Stream): A list of river.Stream objects containing the
            streaming data to be evaluated. If a single river.Stream object is provided,
            it is automatically converted to a list.
        metric (river.metrics.base.MultiClassMetric or river.metrics.base.RegressionMetric):
            The metric to be used for evaluation.
        models (dict): A dictionary of OML models to be evaluated. The keys are the names
            of the models, and the values are the model objects.
        step (int): Iteration number at which to yield results. This only takes into account
            the predictions, and not the training steps. Defaults to 100.
        weight_coeff (float): Results are multiplied by (step/n_steps)**weight_coeff,
            where n_steps is the total number of iterations. Results from the beginning have
            a lower weight than results from the end if weight_coeff > 1. If weight_coeff == 0,
            then results are multiplied by 1 and every result has an equal weight. Defaults to 0.0.
        log_level (int): The level of logging to use. 0 = no logging, 50 = print only important
            information. Defaults to 50.

    Returns:
        (dict): A dictionary containing the evaluation results. The keys are the names of the
            models, and the values are dictionaries with the following keys:
            - "step": A list of iteration numbers at which the model was evaluated.
            - "error": A list of the weighted errors for each iteration.
            - "r_time": A list of the weighted running times for each iteration.
            - "memory": A list of the weighted memory usages for each iteration.
            - "metric_name": The name of the metric used for evaluation.

    Reference:
        https://riverml.xyz/0.15.0/recipes/on-hoeffding-trees/

    Examples:
        >>> from river import compose
            from river import linear_model
            from river import preprocessing, datasets, utils, metrics
            import matplotlib.pyplot as plt
            from spotRiver.utils.features import get_ordinal_date
            from spotRiver.evaluation.eval_nowcast import eval_nowcast_model, plot_nowcast_model
            model = compose.Pipeline(
                ('ordinal_date', compose.FuncTransformer(get_ordinal_date)),
                ('scale', preprocessing.StandardScaler()),
                ('lin_reg', linear_model.LinearRegression())
            )
            dataset = datasets.AirlinePassengers()
            dates, metric, y_trues, y_preds = eval_nowcast_model(model, dataset=dataset)
            plot_nowcast_model(dates, metric, y_trues, y_preds)

    """
    metric_name = metric.__class__.__name__
    # Convert dataset to a list if needed
    if dataset.__class__ != list:
        dataset = [dataset]
    n_steps = len(dataset)
    result = {}
    for model_name, model in models.items():
        result_i = {"step": [], "error": [], "r_time": [], "memory": []}
        for checkpoint in iter_progressive_val_score(
            dataset, model, metric, measure_time=True, measure_memory=True, step=step
        ):
            if log_level <= 20:
                progress_bar(checkpoint["Step"] / n_steps, message="Eval iter_prog_val_score:")
            w = (checkpoint["Step"] / n_steps) ** weight_coeff
            result_i["step"].append(checkpoint["Step"])
            result_i["error"].append(w * checkpoint[metric_name].get())
            # Convert timedelta object into seconds
            result_i["r_time"].append(w * checkpoint["Time"].total_seconds())
            # Make sure the memory measurements are in MB
            raw_memory = checkpoint["Memory"]
            result_i["memory"].append(w * raw_memory * 2**-20)
        result_i["metric_name"] = metric_name
        result[model_name] = result_i
    return result


def plot_oml_iter_progressive(result, log_x=False, log_y=False, figsize=None, filename=None):
    """Plot evaluation of OML models.

    Args:
        result (dict): A dictionary of evaluation results, as returned by eval_oml_iter_progressive.
        log_x (bool, optional): If True, the x-axis is set to log scale. Defaults to False.
        log_y (bool, optional): If True, the y-axis is set to log scale. Defaults to False.
        figsize (tuple, optional): The size of the figure. Defaults to None, in which case
            the default figure size `(10, 5)` is used.
        filename (str, optional): The name of the file to save the plot to. If None, the plot
            is not saved. Defaults to None.

    Returns:
        (matplotlib.figure.Figure): The figure object.

    Reference:
        https://riverml.xyz/0.15.0/recipes/on-hoeffding-trees/

    Examples:
        >>> from spotRiver.evaluation.eval_oml import plot_oml_iter_progressive

        >>> result = {
        ...     "model1": {
        ...         "step": [1, 2, 3],
        ...         "error": [0.1, 0.2, 0.3],
        ...         "r_time": [0.1, 0.2, 0.3],
        ...         "memory": [0.1, 0.2, 0.3],
        ...         "metric_name": "MAE"
        ...     },
        ...     "model2": {
        ...         "step": [1, 2, 3],
        ...         "error": [0.2, 0.3, 0.4],
        ...         "r_time": [0.2, 0.3, 0.4],
        ...         "memory": [0.2, 0.3, 0.4],
        ...         "metric_name": "MAE"
        ...     }
        ... }

        >>> plot_oml_iter_progressive(result)
        <Figure size 1000x500 with 3 Axes>

    """
    if figsize is None:
        figsize = (10, 5)
    fig, ax = plt.subplots(figsize=figsize, nrows=3, dpi=300)
    for model_name, model in result.items():
        ax[0].plot(model["step"], model["error"], label=model_name)
        ax[1].plot(model["step"], model["r_time"], label=model_name)
        ax[2].plot(model["step"], model["memory"], label=model_name)

    ax[0].set_ylabel(model["metric_name"])
    ax[1].set_ylabel("Time (seconds)")
    ax[2].set_ylabel("Memory (MB)")
    ax[2].set_xlabel("Instances")

    ax[0].grid(True)
    ax[1].grid(True)
    ax[2].grid(True)
    if log_y:
        ax[0].set_yscale("log")
        ax[1].set_yscale("log")
        ax[2].set_yscale("log")

    ax[0].legend(loc="upper center", bbox_to_anchor=(0.5, 1.25), ncol=3, fancybox=True, shadow=True)
    plt.tight_layout()
    plt.close()
    if filename is not None:
        fig.savefig(filename, dpi=300)
    return fig


def fun_eval_oml_iter_progressive(result, metric=None, weights=None):
    """Wrapper function for eval_oml_iter_progressive, returning a single function value.

    Args:
        result (dict): A dictionary of evaluation results, as returned by eval_oml_iter_progressive.
        metric (function, optional): The metric function to use for computing the function value.
            Defaults to None, in which case the mean function is used.
        weights (numpy.array, optional): An array of weights for error, r_time, and memory.
            If None, the weights are set to [1, 0, 0], which considers only the error.
            Defaults to None.

    Returns:
        (numpy.array): An array of function values, one for each model in the evaluation results.

    Raises:
        ValueError: If the weights array is not of length 3.

    Reference:
        https://riverml.xyz/0.15.0/recipes/on-hoeffding-trees/

    Examples:
        >>> from spotRiver.evaluation.eval_oml import fun_eval_oml_iter_progressive

        >>> result = {
        ...     "model1": {
        ...         "step": [1, 2, 3],
        ...         "error": [0.1, 0.2, 0.3],
        ...         "r_time": [0.1, 0.2, 0.3],
        ...         "memory": [0.1, 0.2, 0.3],
        ...         "metric_name": "MAE"
        ...     },
        ...     "model2": {
        ...         "step": [1, 2, 3],
        ...         "error": [0.2, 0.3, 0.4],
        ...         "r_time": [0.2, 0.3, 0.4],
        ...         "memory": [0.2, 0.3, 0.4],
        ...         "metric_name": "MAE"
        ...     }
        ... }

        >>> fun_eval_oml_iter_progressive(result)
        array([0.1, 0.2])

    """
    if metric is None:
        metric = mean
    if weights is None:
        weights = array([1, 0, 0])
    if len(weights) != 3:
        raise ValueError("The weights array must be of length 3.")
    model_names = list(result.keys())
    n = len(model_names)
    y = zeros([n])
    for i in range(n):
        y[i] = (
            weights[0] * metric(result[model_names[i]]["error"])
            + weights[1] * metric(result[model_names[i]]["r_time"])
            + weights[2] * metric(result[model_names[i]]["memory"])
        )
    return y
